library(shiny)
library(spsComps)

#fn: color transparency
t_col<-function(color, percent = 50, name = NULL) {
  rgb.val <- col2rgb(color)
  t.col <- rgb(rgb.val[1], rgb.val[2], rgb.val[3],
               max = 255,
               alpha = (100-percent)*255/100,
               names = name)
  invisible(t.col)
}	


#fn: consensus by time window
fn_consensus_by_time_window<-function(seqs,time_windows,ref,name,protein,fn_consensus_by_time_window_dir,systime){
  
  
  #fn: how to compute consensus
  tie_order<-c("L","G","I","K","E","A","V","T","R","Q","S","P","N","D","W","Y","F","H","C","M","-")
  consensusBase_freq<-function(site,threshold){
    site<-toupper(site)
    site <- site[site!="-"] ## exclude gaps
    lth <- length(site)
    if (max(table(site))>lth*threshold){
      base.call<-names(table(site))[(table(site)==max(table(site)))]
      if(length(base.call)>1){
        base.call<-base.call[order(match(toupper(base.call),tie_order))]
        base.call<-base.call[1]
      }
    } else base.call <- "-"
    return(base.call)
  }
  
  
  #set subtype/CRF colors, pch
  subtype_cols<-setNames(viridis::turbo(25),c('01_AE','01B','01BC','02_AG','02A','02A1','07_BC','08_BC','107','11_cpx','35_A1D','63_02A6','A1','A1C','A1CD','A1D','A6','B','BC','BF1','C','D','F1','G','O'))
  set_cols<-subtype_cols[which(names(subtype_cols)%in%name)]
  pch_div<-c(0,1,2,5,6)[1:length(time_windows)]
  pch_pd<-c(22,21,24,23,25)[1:length(time_windows)]
  
  #set reference sequence
  refseq<-ref[grep(protein,rownames(ref)),]
  refseq<-refseq[which(refseq!='-')]
  
  #Ab contact sites
  envf<-read.csv('data/env_features_all.csv',header=T)
  contacts<-function(start,stop){
    c()->x
    for(i in start:stop){
      c(x,grep(1,envf[,i]))->x}
    return(unique(sort(x)))}
  cd4<-grep(1,envf[,'cd4_contacts'])
  v1v2<-contacts(27,30)
  v3<-contacts(31,34)
  mper<-c(660:683)
  eps<-c('vrc01_contacts','cap256','pgt121','x10e8_contacts','x35O22_contacts')
  contact_names<-c('CD4bs','V1V2','V3','MPER','VRC01','CAP256','PGT121','10E8','35O22')
  ep_sites<-lapply(eps,function(e) which(envf[,which(colnames(envf)==e)]==1))
  contact_sites<-setNames(list(cd4,v1v2,v3,mper,ep_sites[[1]],ep_sites[[2]],ep_sites[[3]],ep_sites[[4]],ep_sites[[5]]),contact_names)
  
  #Drug-resistant mutation sites
  enzyme_fas<-read.fasta('data/enzymes.fas')$ali
  enz_hxb2<-enzyme_fas[,which(enzyme_fas['HXB2',]%in%LETTERS)]
  enzymes<-c('PR','RT','INT')
  pro_drm<-c(30,32,46:48,50,54,76,82,84,88,90)
  rt_drm<-c(41,65,67:70,74,115,151,184,210,215,100,101,
            103,106,138,181,188,190,225,227,230)
  int_drm<-c(66,92,118,121,138,140,143,145:148,151,155,263)
  drms<-setNames(list(pro_drm,rt_drm,int_drm),enzymes)  
  
  #fn: compute consensuses, non-consensus residues
  fn_con_difs_over_time<-function(o,protein){
    
    ##metadata
    o<-as.matrix(o)
    dat<-as.data.frame(str_split_fixed(rownames(o),'\\.',n=5))
    colnames(dat)<-c('subtype','country','year','patient','accession')
    
    ##consensus residues (%) by time window
    l=ncol(o)
    cons<-setNames(lapply(time_windows,function(y){
      w<-which(dat$year>=y[1] & dat$year<=y[2])
      
      if(length(w)<5) seq_cons<-rep('X',l)
      else{
        f_time<-as.character(as.matrix((o[w,])))
        
        #remove sites with >50% gaps
        seqs_nogap<-f_time[,which(colMeans(f_time!="-") > 0.5)]
        gaps_removed=ncol(f_time)-ncol(seqs_nogap)
        
        #compute consensus, align to sequences
        seq_cons<-apply(seqs_nogap,2,consensusBase_freq,threshold=0)	
        f_cons<-align(seq_cons,f_time)
        con_row<-grep('newyrow',rownames(f_cons))
        rownames(f_cons)[con_row]<-paste0(dat$subtype[1],'_consensus_',y[1])
        rem_con_gaps<-which(colMeans(f_cons=='-')==1)
        if(length(rem_con_gaps)==0) f_cons_nogap<-f_cons
        else f_cons_nogap<-f_cons[,-which(colMeans(f_cons=='-')==1)]
        
        #percentage of consensus residues per site
        cons_per_site<-apply(f_cons_nogap,2,function(i) 
          round(length(which(i[-con_row]==i[con_row]))/nrow(f_cons_nogap),3)*100)
        
        list(seq_cons=seq_cons,cons_per_site=cons_per_site)
      }
    }),sapply('T',paste0,seq_along(time_windows)))
    
    #align consensus sequences
    these_cons<-which(lengths(cons)==2)
    
    if(length(these_cons)==0) {
      shinyjs::enable("Submit")
      text<-paste0('Not enough ',dat[1,1],' sequences in time window(s)')
      shinyCatch(message(text),position="bottom-right",blocking_level="message")
    }

    cons_al<-refseq
    for(a in these_cons){
      cons_al<-align(cons_al,cons[[a]]$seq_cons)
    }
    rownames(cons_al)<-c('Ref',sapply(time_windows[these_cons],function(i) paste0(dat$subtype[1],'_',i[1],'-',i[2])))
    
    
    ##mismatches between consensuses
    #if only one consensus sequence, return NA
    if(nrow(cons_al)<3) dif_cons<-NA
    
    #if >1 consensus sequence
    else{	
      samples<-combn(1:nrow(cons_al[-1,]),2)
      dif_cons<-apply(samples,2,function(j){
        cons_al_hxb2<-cons_al[,which(cons_al[1,]%in%LETTERS)]
        apply(cons_al_hxb2[j+1,],2,function(k) length(table(k)))
      })
      colnames(dif_cons)<-paste0('T',apply(combn(these_cons,2),2,paste,collapse=''))
      rownames(dif_cons)<-1:length(refseq)	
    }
    
    ##Sites of interest
    #Env - Ab contact sites
    if(protein=='Env'){
      #if only one consensus sequence, return NA
      if(nrow(cons_al)<3) dif_cons_ab<-NA
      
      #if >1 consensus sequence
      else{
        dif_cons_ab<-apply(dif_cons,2,function(l){
          difs<-which(l>1)
          setNames(sapply(contact_sites,intersect,difs),names(contact_sites))
        })
      }
      dif_cons_drm<-NA
    }	
    
    #Pol - drug-resistant mutations	
    if(protein=='Pol'){
      cons_al_hxb2<-cons_al[,which(cons_al['Ref',]%in%LETTERS)]
      enz_cons<-rbind(enz_hxb2,cons_al_hxb2)
      dif_cons_drm<-sapply(enzymes,function(g){
        x<-grep(g,rownames(enz_cons))
        drm<-drms[[grep(g,names(drms))]]
        sapply(time_windows,function(h){
          y<-grep(h[1],rownames(enz_cons))
          if(length(y)==0) NA
          else{
            z<-enz_cons[c(x,y),]
            drmz<-z[,which(z[1,]!='-')[drm]]
            sum(apply(drmz,2,function(i) length(table(i)))-1)
          }
        })
      })  
      dif_cons_ab<-NA
    }
    
    if(protein!='Env' | protein!='Pol'){
      dif_cons_ab<-NA
      dif_cons_drm<-NA
    }	
    
    res<-list(cons_alignment=cons_al,consensus_data=cons,
              consensus_differences=dif_cons,
              ab_consensus_differences=dif_cons_ab,
              drm_consensus_differences=dif_cons_drm)
  }
  
  ##Run fn_con_difs_over_time
  
  ##setup plotting window
  #if sites of interest to be plotted
  if(protein=='Env' | protein=='Pol'){
    par(mar=c(5,5,1,5),xpd=T)
    layout(matrix(c(1,2,3,3),2,2,byrow=T))
  }
  
  #if no sites of interest to be plotted
  else par(mfrow=c(1,2),mar=c(5,5,1,5),xpd=T)	
  
  #if only one subtype/CRF
  if(length(seqs)==1){
    con_difs_over_time<-fn_con_difs_over_time(seqs[[1]],protein=protein)
    consensus_alignments<-as.AAbin(con_difs_over_time$cons_alignment)
    xlim<-(length(time_windows)+1)
    times<-paste0('T',1:length(time_windows))
    
    #concatenate non-consensus residues
    tab<-data.frame(sapply(times,function(s){
      g<-grep(s,names(con_difs_over_time$consensus_data))
      h<-con_difs_over_time$consensus_data[[g]]
      if(length(h)==2) quantile(h$cons_per_site,na.rm=T)
      else rep(NA,5)
    })	)			
    rownames(tab)<-c('min','iqr1','med','iqr3','max')
    
    #plot non-consensus residues 
    plot(as.numeric(tab['med',]),
         xlim=c(0,length(times)),ylim=c(0,100),ann=F,axes=F,
         pch=pch_div,col=set_cols,cex=2)
    mtext('Consensus residues (%)',2,2,cex=0.75)
    axis(2,las=2,mgp=c(3,0.5,0))
    axis(1,at=1:length(times),labels=times)
    suppressWarnings({
      arrows(1:length(times),as.numeric(tab['min',]),
             1:length(times),as.numeric(tab['max',]),
             length=0.05,angle=90,code=3,lty=3,
             col=set_cols)
      arrows(1:length(times),as.numeric(tab['iqr1',]),
             1:length(times),as.numeric(tab['iqr3',]),
             length=0.05,angle=90,code=3,
             col=set_cols)
    })
    legend('topright',inset=c(-0.3,0),sapply(time_windows,function(p) paste0(p[1],'-',p[2])),pch=pch_div,bty='n',cex=0.75)
    
    ##consensus differences between time windows
    nr<-nrow(con_difs_over_time$cons_alignment)
    
    #if only one consensus sequence
    if(nr<=2){
      con_dif_counts<-0
      ymax=1
      ts<-'T1'
    }
    
    #if >1 consensus sequence
    else {
      sams<-combn(1:nrow(con_difs_over_time$cons_alignment[-1,]),2)
      ts<-paste0('T',apply(sams,2,paste0,collapse=''))
      con_dif_counts<-apply(con_difs_over_time$consensus_differences-1,2,sum)
      ymax=max(con_dif_counts,na.rm=T)
    }
    
    #plot consensus differences between time windows
    b<-barplot(con_dif_counts,col=sapply(set_cols,t_col,seq(0,100,100/length(ts))[1:length(ts)]),beside=T,ann=F,axes=F,ylim=c(0,ymax))
    axis(2,las=2,mgp=c(3,0.5,0))	
    mtext('Consensus mismatches',2,2.2,cex=0.75)
    legend('topright',inset=c(-0.2,0),ts,col=sapply(1,t_col,seq(0,100,100/length(ts))[1:length(ts)]),bty='n',pch=15)
    
    ##Sites of interest
    
    #if protein is Env or Pol
    if(protein=='Env' | protein=='Pol') {
      
      #if protein is Env
      if(protein=='Env'){
        
        ylab='Mismatched consensus sites'
        cnames<-contact_names
        
        #if only one consensus sequence
        if(sum(lengths(con_difs_over_time$ab_consensus_differences))==1){
          sois_tabs<-rep(0,9)
          names(sois_tabs)<-cnames
          t_cols<-1
          namesarg<-cnames
          xaxisgo<-0
        }
        
        #if >1 consensus sequence
        else{
          sois_tabs<-sapply(con_difs_over_time$ab_consensus_differences,lengths)
          t_cols<-sapply(set_cols,t_col,seq(0,100,100/length(ts))[1:length(ts)])
          namesarg<-rep(cnames,length(ts))
          xaxisgo<-length(ts)
          xaxis<-ts
        }
      }
      
      #if protein is Pol
      if(protein=='Pol'){
        
        ylab='Consensus sites mutated wrt DRMs'
        sois_tabs<-t(con_difs_over_time$drm_consensus_differences)
        t_cols<-sapply(set_cols,t_col,seq(0,100,100/length(ts))[1:length(ts)])
        cnames<-names(drms)
        if(length(time_windows)>1) namesarg<-rep(cnames,length(time_windows))
        else namesarg<-cnames
        xaxisgo<-length(time_windows)
        xaxis<-sapply(time_windows,function(w) paste0(w[1],'-',w[2]))
      }
      
      #max y-axis value 
      ymax<-max(sois_tabs,na.rm=T)
      ymax<-ifelse(ymax>0,ymax,1)  
      
      #plot sites of interest
      b<-barplot(sois_tabs,col=set_cols,cex.names=0.5,las=2,mgp=c(3,0.5,0),beside=T,ylim=c(0,ymax),names.arg=namesarg)
      mtext(ylab,2,2,cex=0.75)	
      if(xaxisgo>1){
        axis(1,at=b[median(1:nrow(b)),],xaxis,pos=-0.1,tick=F,line=0)  
      }
      
    }
    
    #if protein is neither Env nor Pol
    else sois_tabs<-NA
    
    #save image
    imagefile<-paste0(fn_consensus_by_time_window_dir,'/consensus_by_time_window_',systime,'.tiff')
    agg_tiff(file = imagefile,width=8,height=10,units="in",res=300)
    dev.set(which = 2)
    dev.copy(which = 4)
    dev.off()
  }
  
  #if >1 subtype/CRF
  else{
    con_difs_over_time<-setNames(lapply(seqs,fn_con_difs_over_time,protein=protein),name)
    consensus_alignments<-lapply(con_difs_over_time,function(w) as.AAbin(w$cons_alignment))
    xlim<-length(seqs)*(length(time_windows)+1)
    l<-round(length(seqs)/2)
    sequ1<-seq(1,xlim,(length(time_windows)+1))	
    sequ2<-seq(length(time_windows),xlim,(length(time_windows)+1))	
    times<-paste0('T',1:length(time_windows))
    
    ##plot consensus residues (%)
    plot(1,1,pch=NA,xlim=c(0,xlim),ylim=c(0,100),ann=F,axes=F)
    axis(1,at=seq(median(1:length(time_windows)),xlim,(length(time_windows)+1)),labels=names(set_cols))
    axis(2,las=2,mgp=c(3,0.5,0))
    mtext('Consensus residues (%)',2,2,cex=0.75)
    
    #concatenate non-consensus residues
    con_res_tabs<-lapply(seq_along(con_difs_over_time),function(q){
      tab<-data.frame(sapply(times,function(s){
        g<-grep(s,names(con_difs_over_time[[1]]$consensus_data))
        h<-con_difs_over_time[[q]]$consensus_data[[g]]
        if(length(h)==2) quantile(h$cons_per_site,na.rm=T)
        else rep(NA,5)
      })	)			
      rownames(tab)<-c('min','iqr1','med','iqr3','max')
      
      #plot points
      points(sequ1[q]:sequ2[q],tab['med',],pch=pch_div,col=set_cols[q],cex=2)
      suppressWarnings({
        arrows(sequ1[q]:sequ2[q],as.numeric(tab['min',]),
               sequ1[q]:sequ2[q],as.numeric(tab['max',]),
               length=0.05,angle=90,code=3,lty=3,col=set_cols[q])
        arrows(sequ1[q]:sequ2[q],as.numeric(tab['iqr1',]),
               sequ1[q]:sequ2[q],as.numeric(tab['iqr3',]),
               length=0.05,angle=90,code=3,col=set_cols[q])	
      })
    })	
    legend('topright',inset=c(-0.2,0),sapply(time_windows,function(p) paste0(p[1],'-',p[2])),
           pch=pch_div,bty='n',cex=0.75)
    
    ##consensus differences between time windows
    
    #if more than one time window
    if(length(time_windows)!=1){
      sams<-lapply(con_difs_over_time,function(u){
        nr<-nrow(u$cons_alignment)
        if(nr==2) NA
        else combn(1:nrow(u$cons_alignment[-1,]),2)	
      })
      #ts<-paste0('T',apply(sams[[which.max(sapply(sams,length))]],2,paste,collapse=''))
      ts<-apply(combn(1:length(time_windows),2),2,function(w) paste0('T',paste0(w,collapse='')))
      con_dif_counts<-sapply(con_difs_over_time,function(t){
        if(length(t$consensus_differences)==1) rep(0,length(ts))
        else{
          difs<-apply(t$consensus_differences,2,function(s) length(which(s>1)))
          difs[match(ts,names(difs))]
        }
      })
      
      #set colors for barplot
      if(length(time_windows)>2) cols=sapply(set_cols,t_col,seq(0,100,100/length(ts))[1:length(ts)])
      else cols=set_cols
      
      #plot consensus differences between time windows
      b<-barplot(con_dif_counts,col=cols,beside=T,ann=F,axes=F,names.arg=name)
      axis(2,las=2,mgp=c(3,0.5,0))	
      mtext('Consensus mismatches',2,2.2,cex=0.75)
      legend('topright',inset=c(-0.2,0),ts,col=sapply(1,t_col,seq(0,100,100/length(time_windows))[1:length(time_windows)]),	bty='n',pch=15)
    }
    
    #if only one time window
    else{
      barplot(0,ylim=c(0,1))
      ts<-'T1'
    }
    
    
    ##Sites of interest
    
    #if protein is Env or Pol
    if(protein=='Env' | protein=='Pol'){
      
      nu<-setNames(lapply(seq_along(con_difs_over_time),function(z){
        
        #if protein is Env
        if(protein=='Env'){
          
          #if no differences between consensuses at sites of interest
          if(sum(lengths(con_difs_over_time[[z]]$ab_consensus_differences))==1){
            sois<-rep(0,9)
            t_cols<-1
          }
          
          #if yes differences between consensuses at sites of interest
          else{
            sois<-t(sapply(con_difs_over_time[[z]]$ab_consensus_differences,lengths))
            rownames(sois)<-paste0(name[z],'.',rownames(sois))
          }
        }	
        
        #if protein is Pol
        if(protein=='Pol'){	
          sois<-t(con_difs_over_time[[z]]$drm_consensus_differences)
          rownames(sois)<-paste0(name[z],'.',rownames(sois))
        }	
        sois
      }),name)
      
      #concatenate differences into table
      c()->sois_tabs
      for(h in 1:length(nu)){
        sois_tabs<-rbind(sois_tabs,nu[[h]])
      }
      
      #set colors, labels
      if(protein=='Env'){
        cnames=contact_names
        ls<-length(ts)
        barcols<-rep(set_cols,each=length(cnames)*ls)
        barnames<-rep(NA,ncol(sois_tabs))
        rowlabels<-rownames(sois_tabs)
        ylab='Mismatched consensus sites'
        #colnames(sois_tabs)<-ts
      }
      if(protein=='Pol'){
        cnames=names(drms)
        ls<-length(time_windows)
        barcols<-rep(set_cols,each=length(cnames)*ls)
        barnames<-rep(NA,nrow(sois_tabs))
        rowlabels<-sapply(1:length(time_windows),function(p) paste0('T',p))
        ylab='Consensus sites mutated wrt DRMs'
        if(ls>1) colnames(sois_tabs)<-rowlabels
      }
      
      #max y-axis value
      ymax<-max(sois_tabs,na.rm=T)
      ymax<-ifelse(ymax>0,ymax,1)
      
      #plot consensus differences at Ab contact sites
      #or consensus sites mismatched with DRM references
      b<-barplot(t(sois_tabs),col=barcols,beside=T,axes=F,mgp=c(3,3,1),ylim=c(0,ymax),cex.names=0.5)
      if(protein=='Pol'){
        lab<-sapply(1:ncol(b),function(q){
          if(ls==1) axis(1,at=b[,q],labels=colnames(sois_tabs),mgp=c(3,0,0),tick=F,cex.axis=0.7)
          else axis(1,at=b[,q],rowlabels,mgp=c(3,0,0),tick=F,cex.axis=0.5,las=2)
        })
      }
      if(protein=='Env'){
        lab<-sapply(1:ncol(b),function(p){
          if(ls==1) axis(1,at=b[,p],labels=colnames(sois_tabs),mgp=c(3,0,0),tick=F,cex.axis=0.7)
          else axis(1,at=b[,p],cnames,mgp=c(3,0,0),tick=F,cex.axis=0.5,las=2)
        })
      }
      mtext(ylab,2,2.2,cex=0.75)
      axis(2,las=2,mgp=c(3,0.5,0))
      #if(protein=='Pol') axis(1,at=b[seq(2,nrow(b),length(cnames)),],labels=rep(rowlabels,2),tick=F,pos=-0.75)
    }
    
    #if protein is neither Env nor Pol  
    else sois_tabs<-NA    
    
    
    imagefile<-paste0(fn_consensus_by_time_window_dir,'/consensus_by_time_window_',systime,'.tiff')
    agg_tiff(file = imagefile,width=4*length(seqs),height=10,units="in",res=300)
    dev.set(which = 2)
    dev.copy(which = 4)
    dev.off()  
  }
  
  
  return(list(consensus_alignments=consensus_alignments,sois_tabs=sois_tabs))
  
}	



